
namespace eval telnet {
    variable Telnet
    set debug 0
    
    namespace export openSocket binarySend telnetEvent

    source binaryStrings.tcl
    source readincl.tcl

    readIncludeFile Lookups /usr/include/arpa/telnet.h
    set Lookups(UNKNOWN.-1) "Unknown Option"

    proc Debugputs {string} {
	variable debug
	if {$debug} {
	    puts $string
	}
    }
    
    proc openSocket {ipAddress {port 23}} {
	variable Telnet
	set Telnet(sock) [socket $ipAddress $port]
	fconfigure $Telnet(sock) -translation binary -buffering none -blocking 0

	fileevent $Telnet(sock) readable "[namespace current]::readData $Telnet(sock)"
	set Telnet(callbackScript) displayText
    }
    
    # The default proc for doing stuff - just display the info
    proc displayText {{text {}}} {
	puts -nonewline $text
    }

    #
    # A simple readData for socket use
    #
    proc readData {s} {
	variable Telnet
	set n [read $s]
	if {($n <= 0) && [eof $s]} {
	    global done
	    close $s
	    set done 1
	}
	Debugputs "READ: [string length $n] characters"
	if {[string length $n] < 30} {
	    Debugputs "CHARS: [binaryStrings::bin2hex $n]"
	}

	foreach {text oob} [SplitOOB $n] { }

	Debugputs "Back from split: TXT: $text OOB: [binaryStrings::bin2hex $oob]"

	if {[string length $oob] > 0} {
	    foreach cmt $Telnet(comment) {
		puts "COMMENT: $cmt"
	    }
	    binarySend $oob
	}
	$Telnet(callbackScript) $text
    }
    
    proc telnetEvent {script} {
	variable Telnet
        set Telnet(callbackScript) $script
    }

    proc SplitOOB {string} {
	variable Lookups
	variable Telnet

	set text ""
	set oob ""

	# Set empty the commentary.
	set Telnet(comment) ""

	set lst [binaryStrings::bin2hex $string]

	for {set i 0} {$i < [llength $lst]} {incr i} {
	    set l [lindex $lst $i]
	    while {[string match $l "ff"]} {
		set Telnet(comment-temp) ""
		incr i
		set m [lindex $lst $i]
		scan $m %x dec
		incr i
		incr i [$m [lrange $lst $i end] text oob]
		set id [array names Lookups *.$dec]
		Debugputs "NEW M: $m :: $dec :: [array names Lookups *.$dec]\
		  :: $oob"
		lappend Telnet(comment) "$Lookups($id): $Telnet(comment-temp)"
		set l [lindex $lst $i]
		Debugputs "NEW I: $i :: NEW-L $l"
	    }
	    append text [string trim $l]
	}
	return [list [binaryStrings::hex2bin $text]\
	  [binaryStrings::hex2bin $oob]]
    }

    proc binarySend {str} {
	variable Telnet
	if {[string length $str] == 0} {return}
	Debugputs "SENDING: [binaryStrings::bin2hex $str]"

	puts -nonewline $Telnet(sock) $str
	flush $Telnet(sock)
    }

    #
    # The procedures for dealing with individual IAC options
    #

    proc f0 {subl textName oobName} {
	error "End of Negotiations - This should not happen"
	return 1
    }

    # Receive a NOOP - ignore it.
    proc f1 {} {return 1 }

    # Receive a Data Mark - ignore it.
    proc f2 {} { return 1 }

    # Receive a Break - ignore it.
    proc f3 {} { return 1 }

    # Receive a Interrupt - ignore it.
    proc f4 {} { return 1 }

    # Receive a Abort output - ignore it.
    proc f5 {} { return 1 }

    # Are you there - send a NO-OP just to say hello.
    proc f6 {subl textName oobName} {
	upvar $oobName oob
	append oob fff1
    }

    # Receive a Erase character - ignore it.
    proc f7 {} { return 1 }

    # Receive a Erase line - ignore it.
    proc f8 {} { return 1 }

    # Receive a go ahead - ignore it.
    proc f9 {} { return 1 }

    # Do the fa Negotiation
    # With checking
    proc fa {subl text oobName} {
	variable Telnet
	upvar $oobName oob
	# Starts with character after the 'fa'
	Debugputs "Dealing with: [lrange $subl 0 20]"
	set count 1
	set type [lindex $subl 0]
	set action [lindex $subl 1]
	foreach {p2 data} [FindFFF0 $subl] {}
        
	if {[lsearch [namespace children ::[namespace current]] *$type.TELOPT] < 0} {
	    error "Unknown telnet sub-option: $type"
	} else {
	    append oob [$type.TELOPT::$action $data]
	    AddInfo $subl TELOPT_
	    return $p2
	}
    }
    
    # Trivial version - let the errors fall where they will

    proc fa {subl text oobName} {
	variable Telnet
	upvar $oobName oob
	# Starts with character after the 'fa'
	Debugputs "Dealing with: [lrange $subl 0 20]"
	set count 1
	set type [lindex $subl 0]
	set action [lindex $subl 1]
	foreach {p2 data} [FindFFF0 $subl] {}
        
        append oob [$type.TELOPT::$action $data]
        AddInfo $subl TELOPT_
        return $p2
    }


    # WILL command fb : Reply DO(fd) or DONT(fe)
    proc fb {subl textName oobName} {
	variable Telnet
	upvar $oobName oob
        
	set opt [lindex $subl 0]
	AddInfo $subl TELOPT_
	append oob [eval $opt.TELOPT::fb]

	return 1
    }

    # WONT(fc) - MUST REPLY DONT (fe)
    proc fc {subl textName oobName} {
	variable Telnet
	upvar $oobName oob

	set opt [lindex $subl 0]
	AddInfo $subl TELOPT_
	append oob [eval $opt.TELOPT::fc]

	return 1
    }
    
    set baseClass { variable clientValue %s supported %s;}

    foreach cant {00 05 21 22 23 24 25 26 27 } {
        namespace eval $cant.TELOPT [format $baseClass "" 0]
        namespace eval $cant.TELOPT [format "proc fb {} {return fffe%s}" $cant]
        namespace eval $cant.TELOPT [format "proc fd {} {return fffc%s}" $cant]
        namespace eval $cant.TELOPT  "proc 01 {} {return {}}"
        namespace eval $cant.TELOPT [format \
         {proc fe {} {variable supported; 
           if {$supported} {
	       return fffc%s
	   } else {
	       set supported 0; 
	       return {}}
         }} $cant]
    
    }

    foreach {can initialVal} {18 "dumb" 
                              1f "0x00500018" 
			      20 "57000,57000" 
			      03 "01" 
			      01 "01"} {
        namespace eval $can.TELOPT [format $baseClass $initialVal 1]
        namespace eval $can.TELOPT [format "proc fd {} {return fffb%s}" $can]
        namespace eval $can.TELOPT [format "proc fb {} {return fffd%s}" $can]

        if {[string first 0x $initialVal] == 0} {
            set hex [string range $initialVal 2 end]
        } else {
            binary scan $initialVal H* hex
        }
    
        proc $can.TELOPT::01 {dummy} [format "return fffa%s00%sfff0" $can $hex]

	namespace eval $can.TELOPT {proc 00 {val} {variable serverValue;
	    set serverValue $val}}

        namespace eval $can.TELOPT [format \
         {proc fe {} {variable supported; 
           if {$supported} {
	       return fffc%s
	   } else {
	       set supported 0; 
	       return {}}
         }} $can]
     }

     # DO command fd : Reply WILL(fb) or WONT (fc)
    proc fd {subl textName oobName} {
	variable Telnet
	upvar $oobName oob
	
	set opt [lindex $subl 0]
	AddInfo $subl TELOPT_
	append oob [eval $opt.TELOPT::fd ]

	return 1
    }

    proc AddInfo {opt prefix} {
	variable Telnet
	variable Lookups

	scan $opt %x dec
	set id [array names Lookups ${prefix}*.$dec]
	if {[string match $id ""]} { set id UNKNOWN.-1 }
	Debugputs "ID: ..$id..  DEC: $dec OPT: ..$opt"
	append Telnet(comment-temp) " $Lookups($id)"

    }

    # DONT(fe) - MUST REPLY WONT fc
    # Don't get into a repeat match - after a few iterations, call it
    #  quits.

    proc fe {subl textName oobName} {
	variable Telnet
	upvar $oobName oob

	set opt [lindex $subl 0]
	AddInfo $subl TELOPT_
	append oob [eval $opt.TELOPT::fe]

	return 1
    }

    # Just walk to the end of a negotiation sequence and 
    #  report where it is and what the value is as a list.

    proc FindFFF0 {lst} {
	set iac 0
	set pos 0
	foreach l $lst {
	    incr pos

	    switch $l {
	    ff { set iac 1 }
	    f0 { if {$iac} { return [list $pos [lrange $lst 1 [expr $pos\
	      - 3]]] } }
	    default { set iac 0 }
	    }
	}
	error "NO END CONDITION"
    }

}


proc bgerror {args} {
    global errorInfo
    puts "BGERR: $args \n--\n$errorInfo"
}

# Default site - local loopback.

set site 127.0.0.1
if {$argc > 0} {
    set site $argv
}

telnet::openSocket $site
set done 0
vwait done


